/* config.h.  Generated from config.h.in by configure.  */
/* config.h.in.  Generated from configure.ac by autoheader.  */

/* DATE tag on */
/* #undef DATE */

/* Debug tag on */
#define DEBUG 1

/* EVOLVE tag on */
/* #undef EVOLVE */

/* Define to 1 if you don't have `vprintf' but do have `_doprnt.' */
/* #undef HAVE_DOPRNT */

/* Define to 1 if you have the <float.h> header file. */
#define HAVE_FLOAT_H 1

/* Define to 1 if you have the `floor' function. */
#define HAVE_FLOOR 1

/* Define to 1 if you have the <inttypes.h> header file. */
#define HAVE_INTTYPES_H 1

/* Define to 1 if you have the `m' library (-lm). */
#define HAVE_LIBM 1

/* Define to 1 if you have the `pow' function. */
#define HAVE_POW 1

/* Define to 1 if you have the `rint' function. */
#define HAVE_RINT 1

/* Define to 1 if you have the `sqrt' function. */
#define HAVE_SQRT 1

/* Define to 1 if you have the <stdint.h> header file. */
#define HAVE_STDINT_H 1

/* Define to 1 if you have the <stdio.h> header file. */
#define HAVE_STDIO_H 1

/* Define to 1 if you have the <stdlib.h> header file. */
#define HAVE_STDLIB_H 1

/* Define to 1 if you have the `strchr' function. */
#define HAVE_STRCHR 1

/* Define to 1 if you have the <strings.h> header file. */
#define HAVE_STRINGS_H 1

/* Define to 1 if you have the <string.h> header file. */
#define HAVE_STRING_H 1

/* Define to 1 if you have the `strstr' function. */
#define HAVE_STRSTR 1

/* Define to 1 if you have the <sys/stat.h> header file. */
#define HAVE_SYS_STAT_H 1

/* Define to 1 if you have the <sys/types.h> header file. */
#define HAVE_SYS_TYPES_H 1

/* Define to 1 if you have the <unistd.h> header file. */
#define HAVE_UNISTD_H 1

/* Define to 1 if you have the `vprintf' function. */
#define HAVE_VPRINTF 1

/* MPI tag on */
#define MPI 1

/* Name of package */
#define PACKAGE "phyml"

/* Define to the address where bug reports for this package should be sent. */
#define PACKAGE_BUGREPORT "stephane.guindon@lirmm.fr"

/* Define to the full name of this package. */
#define PACKAGE_NAME "PhyML"

/* Define to the full name and version of this package. */
#define PACKAGE_STRING "PhyML 3.3.20250429"

/* Define to the one symbol short name of this package. */
#define PACKAGE_TARNAME "phyml"

/* Define to the home page for this package. */
#define PACKAGE_URL ""

/* Define to the version of this package. */
#define PACKAGE_VERSION "3.3.20250429"

/* PHYML tag on */
#define PHYML 1

/* PHYREX tag on */
/* #undef PHYREX */

/* PHYREXSIM tag on */
/* #undef PHYREXSIM */

/* PHYTIME tag on */
/* #undef PHYTIME */

/* RF tag on */
/* #undef RF */

/* RWRAP tag on */
/* #undef RWRAP */

/* Define to 1 if all of the C90 standard headers exist (not just the ones
   required in a freestanding environment). This macro is provided for
   backward compatibility; new code need not use it. */
#define STDC_HEADERS 1

/* TEST tag on */
/* #undef TEST */

/* Unix tag on */
#define UNIX 1

/* Version number of package */
#define VERSION "3.3.20250429"

/* WIN32 tag on */
/* #undef WIN32 */

/* Define to empty if `const' does not conform to ANSI C. */
/* #undef const */

/* Define to `__inline__' or `__inline' if that's what the C compiler
   calls it, or to nothing if 'inline' is not supported under any name.  */
#ifndef __cplusplus
/* #undef inline */
#endif

/* Define to `unsigned int' if <sys/types.h> does not define. */
/* #undef size_t */
